/*
 * Funambol is a mobile platform developed by Funambol, Inc. 
 * Copyright (C) 2003 - 2009 Funambol, Inc.
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License version 3 as published by
 * the Free Software Foundation with the addition of the following permission 
 * added to Section 15 as permitted in Section 7(a): FOR ANY PART OF THE COVERED
 * WORK IN WHICH THE COPYRIGHT IS OWNED BY FUNAMBOL, FUNAMBOL DISCLAIMS THE 
 * WARRANTY OF NON INFRINGEMENT  OF THIRD PARTY RIGHTS.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Affero General Public License 
 * along with this program; if not, see http://www.gnu.org/licenses or write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA.
 * 
 * You can contact Funambol, Inc. headquarters at 643 Bair Island Road, Suite 
 * 305, Redwood City, CA 94063, USA, or at email address info@funambol.com.
 * 
 * The interactive user interfaces in modified source and object code versions
 * of this program must display Appropriate Legal Notices, as required under
 * Section 5 of the GNU Affero General Public License version 3.
 * 
 * In accordance with Section 7(b) of the GNU Affero General Public License
 * version 3, these Appropriate Legal Notices must retain the display of the
 * "Powered by Funambol" logo. If the display of the logo is not reasonably 
 * feasible for technical reasons, the Appropriate Legal Notices must display
 * the words "Powered by Funambol".
 */

package com.funambol.blackberry.autotest;

import java.util.Hashtable;
import java.util.Vector;
import java.io.IOException;
import java.io.InputStream;

import com.funambol.syncml.spds.SyncItem;
import com.funambol.syncml.spds.SyncManager;
import com.funambol.syncml.spds.SyncSource;
import com.funambol.syncml.spds.SyncConfig;
import com.funambol.syncml.spds.SyncListener;
import com.funambol.syncml.spds.SyncException;
import com.funambol.syncml.spds.SourceConfig;
import com.funambol.syncml.protocol.SyncFilter;
import com.funambol.syncml.protocol.SyncML;

import com.funambol.client.test.BasicScriptRunner;
import com.funambol.client.test.BasicCommandRunner;
import com.funambol.client.test.ContactsCommandRunner;
import com.funambol.client.test.CalendarCommandRunner;
import com.funambol.client.test.Robot;

import com.funambol.util.Log;
import com.funambol.util.StringUtil;
import com.funambol.util.CodedException;
import com.funambol.util.HttpTransportAgent;
import com.funambol.platform.FileAdapter;

/**
 * Blackberry automatic test suite runner that uses a
 * BlackBerryBasicScriptRunner instance to define all of the CommadRunner
 * objects needed to execute the test suite defined in the given script url.
 */
public class ScriptRunner {

    private static final String SCRIPT_URL_PROPERTY = "ScriptUrl";

    private static final String TEST_NAME = "Script Runner";
    private static final String TAG_LOG   = "ScriptRunner";

    private String details = null;
    private String scriptUrl = null;

    protected BlackBerryBasicScriptRunner basicRunner;
    protected BlackBerryRobot   robot;
    protected BlackBerrySyncMonitor syncMonitor;

    /**
     * Constructor: needs the script url of the test suite scripts
     * @param scriptUrl the String formatted url of the first test of the suite
     */
    public ScriptRunner(String scriptUrl) {
        this();
        this.scriptUrl = scriptUrl;
    }

    /**
     * Empty Constructor. Add the CommandRunners to the command runner container
     * and defines the BlackBerrySyncMonitor and robot related objects.
     */
    public ScriptRunner() {

        Log.setLogLevel(Log.TRACE);
        
        syncMonitor = new BlackBerrySyncMonitor();
        BlackBerryCheckSyncClient checkClient = new BlackBerryCheckSyncClient();

        robot = new BlackBerryRobot();
        robot.setScriptRunner(basicRunner);

        // Create the extended command runner
        basicRunner = new BlackBerryBasicScriptRunner();
        BlackBerryBasicCommandRunner bbBasicRunner = new BlackBerryBasicCommandRunner(robot);
        bbBasicRunner.setSyncMonitor(syncMonitor);
        bbBasicRunner.setCheckSyncClient(checkClient);
        basicRunner.addCommandRunner(bbBasicRunner);

        BlackBerryCommandRunner bbRunner = new BlackBerryCommandRunner(robot);
        bbRunner.setSyncMonitor(syncMonitor);
        bbBasicRunner.setCheckSyncClient(checkClient);
        basicRunner.addCommandRunner(bbRunner);

        BlackBerryContactsRobot contactsRobot = new BlackBerryContactsRobot();
        contactsRobot.setScriptRunner(basicRunner);
        ContactsCommandRunner contactsRunner = new ContactsCommandRunner(contactsRobot);
        contactsRunner.setSyncMonitor(syncMonitor);
        contactsRunner.setCheckSyncClient(checkClient);
        basicRunner.addCommandRunner(contactsRunner);

        BlackBerryCalendarsRobot calendarRobot = new BlackBerryCalendarsRobot();
        calendarRobot.setScriptRunner(basicRunner);
        CalendarCommandRunner calendarRunner = new CalendarCommandRunner(calendarRobot);
        calendarRunner.setSyncMonitor(syncMonitor);
        calendarRunner.setCheckSyncClient(checkClient);
        basicRunner.addCommandRunner(calendarRunner);
    }

    /**
     * Runs the effective test suite invoking the CommandRunner container
     * @throws Throwable
     */
    public void test() throws Throwable {
        //String scriptUrl = FunambolTest.getProperty(SCRIPT_URL_PROPERTY);
        //String scriptUrl = "http://d.funambol.com/Test/federico/bbscript.txt";
        if (scriptUrl == null) {
            scriptUrl = "http://d.funambol.com/Test/carlo/bb/Regression.txt";
        }
        runScriptFile(scriptUrl, true);
    }

    /**
     * Accessor method
     * @return String the detailed content result of the test suite
     */
    public String getDetails() {
        return basicRunner.getResults();
    }

    /**
     * Gets the error code returned after the test suite run
     * @return boolean true if the test suite is passed, false otherwise
     */
    public boolean isTestSuitePassed() {
        return basicRunner.getErrorCode() == 0 ? true : false;
    }

    /**
     * Accessor method
     * @return BlackBerrySyncMonitor the current BlackBerrySyncMonitor object 
     * used by this test suite
     */
    public BlackBerrySyncMonitor getSyncMonitor() {
        return syncMonitor;
    }

    /**
     * Run the script file loading it from the specified url
     * @param scriptUrl the String formatted url of the script file
     * @param mainScript set if the given script file is the main script
     * (not used up to now)
     * @throws Throwable if anything went wrong when loading the script
     */
    private void runScriptFile(String scriptUrl, boolean mainScript) throws Throwable {
        basicRunner.runScriptFile(scriptUrl, mainScript);
    }

    /**
     * General method to tear down the test environment after the tests run.
     * Not yet implemented
     */
    public void tearDown() {
    }

    /**
     * General method to se up the test environment before the tests run.
     * Not yet implemented;
     */
    public void setUp() {
    }
}
 
