/*
 * Funambol is a mobile platform developed by Funambol, Inc. 
 * Copyright (C) 2009 Funambol, Inc.
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License version 3 as published by
 * the Free Software Foundation with the addition of the following permission 
 * added to Section 15 as permitted in Section 7(a): FOR ANY PART OF THE COVERED
 * WORK IN WHICH THE COPYRIGHT IS OWNED BY FUNAMBOL, FUNAMBOL DISCLAIMS THE 
 * WARRANTY OF NON INFRINGEMENT  OF THIRD PARTY RIGHTS.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Affero General Public License 
 * along with this program; if not, see http://www.gnu.org/licenses or write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA.
 * 
 * You can contact Funambol, Inc. headquarters at 643 Bair Island Road, Suite 
 * 305, Redwood City, CA 94063, USA, or at email address info@funambol.com.
 * 
 * The interactive user interfaces in modified source and object code versions
 * of this program must display Appropriate Legal Notices, as required under
 * Section 5 of the GNU Affero General Public License version 3.
 * 
 * In accordance with Section 7(b) of the GNU Affero General Public License
 * version 3, these Appropriate Legal Notices must retain the display of the
 * "Powered by Funambol" logo. If the display of the logo is not reasonably 
 * feasible for technical reasons, the Appropriate Legal Notices must display
 * the words "Powered by Funambol".
 */

package com.funambol.blackberry.autotest;

/**
 * This component lists all the commands available in the automatic test
 * scripting language.
 */
public interface UserCommands {


    /**
     * This command simulates a user clicking on a menu item. The menu is actually opened
     * and the selected voice "clicked".
     *
     * @param item the string shown on the menu item 
     */
    public void ClickMenuItem(String item);

    /**
     * This command waits for a certain screen to be displayed and blocks the
     * script execution until such a screen is visible or the given timeout expires.
     *
     * @param name the screen name is a string that can take one of the values
     * listed here below.
     * <ul>
     *   <li> RimAccountScreen is the screen to set the account. This screen is
     *        used for both the login and any further account change</li>
     *   <li> RimHomeScreen is the main screen of the application</li>
     * </ul>
     * @param timeout the timeout is an integer value expressed in milliseconds.
     */
    public void WaitForScreen(String name, int timeout);

    /**
     * This command simulates a user entering a value into a field.
     * The field is the one that has the "focus". If the focus is on a field for
     * which it does not make sense to enter text, then the command and the test
     * fail.
     * @param value an arbitrary string to be set in the field. If the field
     *              accepts only one type of characters then the value shall
     *              contain only valid characters. If invalid characters are
     *              supplied the test behavior is unpredictable.
     *              For choice groups, the value represent the index (an integer
     *              starting from 0) of the selected item
     */
    public void SetFieldText(String value);

    /**
     * This command suspends the test execution for the given amount of time.
     * @param delay is an integer value expressed in milliseconds. If the delay is <= 0
     *              then the command wait forever and the script gets interrupted.
     *              This is useful if the script shall give the user the possibility to use the application.
     */
    public void Wait(int delay);

    /**
     * This instruction must be the first one at the beginning of a test.
     * For a single test script there is only one of such instruction, while a multi test
     * script may contain several instances.
     * @param name is the name of the test as it will be reported in the final summary.
     */
    public void BeginTest(String name);

    /**
     * Terminates the commands of a test.
     */
    public void EndTest();

    /**
     * This command clears the client configuration. After this command is executed,
     * the client is configuration-less.
     */
    public void ResetConfig();

    /**
     * This command starts the main application. When the test starts the main
     * application (BlackBerry Sync Client) is not started until the script triggers this command.
     */
    public void StartMainApp();

    /**
     * This command wait for a sync to start and finish within a maximum amount
     * of time. As soon as the sync terminates, the script continues its
     * execution
     *
     * @param minStart is the time the script waits for the sync to start. If
     *                 the sync does not start withing this time, the test fails
     * @param maxTime  is the maximum time the script is willing for the sync to
     *                 terminate. If it does not terminate within this limit,
     *                 the test fails.
     */
    public void WaitForSyncToComplete(String minStart, String maxTime);

    /**
     * This command includes another script file.
     *
     * @param name is the script name. If this is an URL, then the script is
     *             fetched from that location, otherwise the base url if the main script is
     *             added as prefix to this name
     */
    public void Include(String name);

    /**
     * This command deletes (resets) all the file in a given directory.
     *
     * @param dir the directory to be reset. Directories in the device memory
     * are named like: file:///store/ while file:///SDCard/ can be used for the
     * memory card.
     */
    public void ResetDir(String dir);

    /**
     * This command copies one file somewhere. File names have the following
     * prefix:
     *
     * <ul>
     *   <li> file:///store/ if they are in the memory of the device </li>
     *   <li> file:///SDCard/ if they are in the sdcard </li>
     * </ul>
     *
     * @param sourceFile the source file name.
     * @param targetFile the target file name.
     */
    public void CopyFile(String sourceFile, String targetFile);

    /**
     * This command renames one file. File names have the following
     * prefix:
     *
     * <ul>
     *   <li> file:///store/ if they are in the memory of the device </li>
     *   <li> file:///SDCard/ if they are in the sdcard </li>
     * </ul>
     *
     * @param sourceFile the source file name.
     * @param targetFile the target file name.
     */
    public void RenameFile(String sourceFile, String targetFile);

    /**
     * This command deletes one file. File names have the following
     * prefix:
     *
     * <ul>
     *   <li> file:///store/ if they are in the memory of the device </li>
     *   <li> file:///SDCard/ if they are in the sdcard </li>
     * </ul>
     *
     * @param file the source file name.
     */
    public void DeleteFile(String file);

    /**
     * This command simulates an user action via the trackwheel.
     * @param command the command to simulate. Possible values are:
     *                <ul>
     *                  <li> Left to roll on the left </li>
     *                  <li> Right to roll on the right </li>
     *                  <li> Down to roll downward </li>
     *                  <li> Up to roll upward </li>
     *                  <li> Click to click (press and release) </li>
     *                </ul>
     * @param repetition this is an optional parameter and it specifies how many
     * times the command shall be repeated. Default is 1.
     */
    public void TrackWheel(String command, int repetition);

    /**
     * This command checks if the name of a source visible in the main screen is
     * a given value.
     *
     * @param idx is the source position (starting from 0 if the SyncAll button
     * is not visible)
     * @param value is the name of the source
     */
    public void CheckVisibleSourceName(int idx, String value);

    /**
     * This command checks if the name of a source visible in the main screen is
     * a given value.
     *
     * @param count is the expected number of sources in the main screen
     * (including, if visible, the SyncAll meta source).
     */
    public void CheckVisibleSourceCount(int count);

    /**
     * This command forces the next sync to be a slow sync for the given source
     *
     * @param sourceName is the name of the source. The value is what is displayed
     * on the main screen for that source.
     */
    public void ForceSlowSync(String sourceName);

    /**
     * This command checks the amount of items exchanged between the server and
     * the client during the last synchronization (for a given source).
     *
     * @param source the source name
     * @param sentAdd the expected number of new items sent
     * @param sentReplace the expected number of replace items sent
     * @param sentDelete the expected number of delete items sent
     * @param receivedAdd the expected number of new received items
     * @param receivedReplace the expected number of replace received items
     * @param receivedDelete the expected number of delete received items
     */
    public void CheckExchangedData(String source, int sentAdd, int sentReplace, int sentDelete,
                                   int receivedAdd, int receivedReplace, int receivedDelete);

    /**
     * This command sets the device timezone to the given value
     *
     * @param tz is the new timezone as available in the UI (e.g.
     * America/Los_Angeles)
     */
    public void SetTimeZone(String tz);


}
    

