/*
 * Funambol is a mobile platform developed by Funambol, Inc.
 * Copyright (C) 2008 Funambol, Inc.
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License version 3 as published by
 * the Free Software Foundation with the addition of the following permission
 * added to Section 15 as permitted in Section 7(a): FOR ANY PART OF THE COVERED
 * WORK IN WHICH THE COPYRIGHT IS OWNED BY FUNAMBOL, FUNAMBOL DISCLAIMS THE
 * WARRANTY OF NON INFRINGEMENT  OF THIRD PARTY RIGHTS.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program; if not, see http://www.gnu.org/licenses or write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA.
 * 
 * You can contact Funambol, Inc. headquarters at 643 Bair Island Road, Suite
 * 305, Redwood City, CA 94063, USA, or at email address info@funambol.com.
 *
 * The interactive user interfaces in modified source and object code versions
 * of this program must display Appropriate Legal Notices, as required under
 * Section 5 of the GNU Affero General Public License version 3.
 *
 * In accordance with Section 7(b) of the GNU Affero General Public License
 * version 3, these Appropriate Legal Notices must retain the display of the
 * "Powered by Funambol" logo. If the display of the logo is not reasonably
 * feasible for technical reasons, the Appropriate Legal Notices must display
 * the words "Powered by Funambol".
 */

package com.funambol.blackberry;

import com.funambol.blackberry.controller.UISyncSourceController;
import com.funambol.blackberry.ui.UISyncSource;
import com.funambol.syncml.spds.SyncListener;
import com.funambol.syncml.spds.SyncSource;
import com.funambol.syncml.spds.SourceConfig;
import com.funambol.syncml.protocol.SyncML;

/**
 * This class represents a sync source in the application. In the application a
 * sync source is a pluggable component that allow users to deal with the
 * synchronization of a given type of data. An AppSyncSource abstracts the
 * following concepts:
 *
 * 1) SyncSource (at SyncML API level)
 * 2) Displayable name
 * 3) Icon (optional)
 * 4) Main configuration screen
 * 5) numerical ID (set by the application and optional)
 * 6) config store id
 * 7) last sync status
 * 8) Position in the UI
 * 9) UI representation (view)
 * 10) UI controller
 *
 */
public class AppSyncSource {

    protected SyncSource           source;
    private String                 name;
    private int                    id     = -1;
    private String                 configStorageKey = null;
    private String                 enabledLabel     = null;
    private String                 disabledLabel    = null;
    private String                 iconName         = null;
    private String                 iconDisabledName = null;
    private ExternalAppManager     appManager       = null;
    private int                    lastSyncStatus   = SyncListener.SUCCESS;
    private int                    uiSourceIndex    = -1;
    private UISyncSource           uiSource         = null;
    private UISyncSourceController controller       = null;
    private boolean                hasSettings      = false;
    protected boolean              refreshFromServerSupported = true;
    protected boolean              refreshToServerSupported   = true;     
    private NativeListener         nativeListener   = null;
    protected String               warningOnFirstSync = null;
    private boolean                syncedInSession  = false;
    
    public AppSyncSource(String name, SyncSource source) {
        this.name   = name;
        this.source = source;
        this.enabledLabel = "";
        this.disabledLabel = "";
    }

    public AppSyncSource(String name) {
        this.name   = name;
        this.source = null;
        this.enabledLabel = "";
        this.disabledLabel = "";
    }

    public void setId(int id) {
        this.id = id;
    }

    public int getId() {
        return id;
    }

    public int getUiSourceIndex() {
        return uiSourceIndex;
    }
    
    public void setUiSourceIndex(int index) {
        this.uiSourceIndex=index;
    }
    
    public void setSyncSource(final SyncSource source) {
        this.source = source;
    }

    public void setAppManager(ExternalAppManager manager) {
        this.appManager = manager;
    }

    public ExternalAppManager getAppManager() {
        return appManager;
    }

    public SyncSource getSyncSource() {
        return source;
    }

    public String getName() {
        return name;
    }

    public String getConfigStorageKey() {
        return configStorageKey;
    }

    public void setConfigStorageKey(String key) {
        configStorageKey = key;
    }

    public String getEnabledLabel() {
        return enabledLabel;
    }

    public void setEnabledLabel(String enabledLabel) {
        this.enabledLabel = enabledLabel;
    }

    public String getDisabledLabel() {
        return disabledLabel;
    }

    public void setDisabledLabel(String disabledLabel) {
        this.disabledLabel = disabledLabel;
    }

    /**
     * This is just a convenience method that fetches information
     * from the Configuration
     */
    public boolean isEnabled() {
        return Configuration.isSourceEnabled(getId());
    }

    public boolean isWorking() {
        return Configuration.isSourceActive(getId()) && source != null;
    }

    public String getIconName() {
        return iconName;
    }

    public void setIconName(String name) {
        iconName = name;
    }

    public void setDisabledIconName(String name) {
        iconDisabledName = name;
    }

    public String getDisabledIconName() {
        return iconDisabledName;
    }

    public void setLastSyncStatus(int status) {
        lastSyncStatus = status;
    }

    public int getLastSyncStatus() {
        return lastSyncStatus;
    }

    public UISyncSource getUISyncSource() {
        return uiSource;
    }

    public void setUISyncSource(UISyncSource uiSource) {
        this.uiSource = uiSource;
    }

    public UISyncSourceController getUISyncSourceController() {
        return controller;
    }

    public void setUISyncSourceController(UISyncSourceController controller) {
        this.controller = controller;
    }

    public boolean hasSettings() {
        return hasSettings;
    }

    public void setHasSettings(boolean value) {
        hasSettings = value;
    }

    public NativeListener getNativeListener() {
        return nativeListener;
    }

    public void setNativeListener(NativeListener nativeListener) {
        this.nativeListener = nativeListener;
    }

    public int prepareRefresh(int direction) throws Exception {
        int syncMode = 0;
        switch (direction) {
            case Constants.REFRESH_FROM_SERVER:
                syncMode = SyncML.ALERT_CODE_REFRESH_FROM_SERVER;
                break;
            case Constants.REFRESH_TO_SERVER:
                syncMode = SyncML.ALERT_CODE_REFRESH_FROM_CLIENT;
                break;
            default:
                throw new IllegalArgumentException("Invalid refresh direction " + direction);
        }
        return syncMode;
    }

    public boolean isRefreshSupported(int direction) {
        boolean supported;
        switch (direction) {
            case Constants.REFRESH_FROM_SERVER:
                supported = refreshFromServerSupported;
                break;
            case Constants.REFRESH_TO_SERVER:
                supported = refreshToServerSupported;
                break;
            default:
                throw new IllegalArgumentException("Invalid refresh direction " + direction);
        }
        return supported;
    }

    public void setIsRefreshSupported(int direction, boolean value) {
        if (direction == Constants.REFRESH_FROM_SERVER) {
            refreshFromServerSupported = true;
        } else if (direction == Constants.REFRESH_TO_SERVER) {
            refreshToServerSupported = true;
        } else {
            throw new IllegalArgumentException("Invalid refresh direction " + direction);
        }
    }

    public void setIsRefreshSupported(boolean value) {
        refreshFromServerSupported = value;
        refreshToServerSupported   = value;
    }

    /**
     * Returns the message to show at the very first sync for this source.
     * The method returns null if there are no messages to show
     */
    public String getWarningOnFirstSync() {
        return warningOnFirstSync;
    }

    public void setWarningOnFirstSync(String message) {
        warningOnFirstSync = message;
    }

    /**
     * This is a transient property that is held across sessions. It indicates
     * if a sync attempt was made for this source in this session.
     */
    public void setSyncedInSession(boolean value) {
        syncedInSession = value;
    }

    public boolean getSyncedInSession() {
        return syncedInSession;
    }

    /**
     * This method is invoked when there is a change in the configuration that
     * impacts the source config. By default the method does not do anything,
     * but the class can be derived to force specific behaviors
     */
    public void reapplyConfiguration() {
    }
}

