/*
 * Funambol is a mobile platform developed by Funambol, Inc.
 * Copyright (C) 2009 Funambol, Inc.
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License version 3 as published by
 * the Free Software Foundation with the addition of the following permission
 * added to Section 15 as permitted in Section 7(a): FOR ANY PART OF THE COVERED
 * WORK IN WHICH THE COPYRIGHT IS OWNED BY FUNAMBOL, FUNAMBOL DISCLAIMS THE
 * WARRANTY OF NON INFRINGEMENT  OF THIRD PARTY RIGHTS.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program; if not, see http://www.gnu.org/licenses or write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA.
 * 
 * You can contact Funambol, Inc. headquarters at 643 Bair Island Road, Suite
 * 305, Redwood City, CA 94063, USA, or at email address info@funambol.com.
 *
 * The interactive user interfaces in modified source and object code versions
 * of this program must display Appropriate Legal Notices, as required under
 * Section 5 of the GNU Affero General Public License version 3.
 *
 * In accordance with Section 7(b) of the GNU Affero General Public License
 * version 3, these Appropriate Legal Notices must retain the display of the
 * "Powered by Funambol" logo. If the display of the logo is not reasonably
 * feasible for technical reasons, the Appropriate Legal Notices must display
 * the words "Powered by Funambol".
 */

package com.funambol.blackberry;

import java.util.Enumeration;
import java.util.Vector;

import javax.microedition.pim.PIMException;
import javax.microedition.pim.PIMItem;
import javax.microedition.pim.PIMList;

import net.rim.blackberry.api.pdap.PIMListListener;
import net.rim.blackberry.api.pdap.BlackBerryPIMList;

import com.funambol.blackberry.AppSyncSource;
import com.funambol.blackberry.source.BBPIMSyncSource;
import com.funambol.blackberry.controller.Controller;
import com.funambol.blackberry.controller.HomeScreenController;

import com.funambol.syncml.spds.SyncException;
import com.funambol.syncml.spds.SyncSource;
import com.funambol.util.Log;

/**
 * Class for PIM listening on BB. This class listens for events on PIM data
 * and notifies the sources so they can track changes and if the application is
 * in push mode a sync is scheduled for the proper source.
 */
public class BBPIMListener implements PIMListListener, NativeListener
{
    protected AppSyncSource        source;
    protected BlackBerryPIMList    list;
    private   AppSyncSourceManager appSyncSourceManager;
    private   Customization        customization;
    private   Configuration        configuration;

    public BBPIMListener(BlackBerryPIMList list, AppSyncSource source, Configuration config)
    {
        this.list   = list;
        this.source = source;
        appSyncSourceManager = AppSyncSourceManager.getInstance();
        customization = Customization.getInstance();
        configuration = config;
    }

    /**
     * This method is a callback invoked when a PIM item is added to the device
     * data base (see PIMListListener for more details).
     * In this implementation the application tracks the new item so that it
     * will be sent in the next sync, and if push is enabled a sync is
     * scheduled.
     * It is important to observe that this method is invoked outside of the
     * application "context". As such it cannot access any static data of the
     * application (for example log messages are not stored in the right place)
     *
     * @param item the PIM item that was added
     */
    public void itemAdded(PIMItem item) {
        // For no reason we want the app to crash here, because we would lose
        // the tracking funcionality completely. Therefore we catch everything.
        Log.trace("[BBPIMListener.itemAdded]");
        try {
            // Notify the source so it can update its tracking data
            BBPIMSyncSource ss = (BBPIMSyncSource)source.getSyncSource();
            ss.itemAddedExternally(item);
            // if we are in push mode, we shall schedule a sync for this source
            programSync();
        } catch (Exception e) {
            Log.error("[BBPIMListener] Error handling itemAdded event " + e.toString());
        }
    }

    /**
     * This method is a callback invoked when a PIM item is removed from the device
     * data base (see PIMListListener for more details).
     * In this implementation the application tracks the removed item so that it
     * will be sent in the next sync, and if push is enabled a sync is
     * scheduled.
     * It is important to observe that this method is invoked outside of the
     * application "context". As such it cannot access any static data of the
     * application.
     *
     * @param item the PIM item that was removed
     */
    public void itemRemoved(PIMItem item) {
        Log.trace("[BBPIMListener.itemRemoved]");
        try {
            BBPIMSyncSource ss = (BBPIMSyncSource)source.getSyncSource();
            ss.itemRemovedExternally(item);
            // if we are in push mode, we shall schedule a sync for this source
            programSync();
        } catch (Exception e) {
            Log.error("[BBPIMListener] Error handling itemRemoved event " + e.toString());
        }
    }

    /**
     * This method is a callback invoked when a PIM item is updated int the device
     * data base (see PIMListListener for more details).
     * In this implementation the application tracks the updated item so that it
     * will be sent in the next sync, and if push is enabled a sync is
     * scheduled.
     * It is important to observe that this method is invoked outside of the
     * application "context". As such it cannot access any static data of the
     * application.
     *
     * @param oldItem the PIM item that was updated (old value)
     * @param newItem the PIM item that was updated (new value)
     */
    public void itemUpdated(PIMItem oldItem, PIMItem newItem) {
        Log.trace("[BBPIMListener.itemUpdated]");
        try {
            BBPIMSyncSource ss = (BBPIMSyncSource)source.getSyncSource();
            ss.itemUpdatedExternally(oldItem, newItem);
            // if we are in push mode, we shall schedule a sync for this source
            programSync();
        } catch (Exception e) {
            Log.error("[BBPIMListener] Error handling itemUpdated event " + e.toString());
        }
    }

    private void programSync() {
        Log.trace("[BBPIMListener.programSync]");
        if (customization.isC2SPushEnabled() && configuration.getC2SPush()) {
            Log.trace("[BBPIMListener] sync mode is push");
            // We always use the home screen controller to fire the sync because
            // we want the sources to be updated in case the screen is shown
            // We must check if this source is enabled before triggering a sync
            Enumeration sources = appSyncSourceManager.getEnabledAndWorkingSources();
            while (sources.hasMoreElements()) {
                AppSyncSource appSource = (AppSyncSource)sources.nextElement();
                if (appSource == source) {
                    // The source is enabled, fire a sync
                    Controller controller = source.getUISyncSourceController().getController();
                    HomeScreenController hsc = controller.getHomeScreenController();

                    // If a sync for this source is already in progress, then
                    // we ignore this change as it is very likely it has been
                    // generated by the sync itself
                    if (!hsc.isSynchronizing() || hsc.getCurrentSource() != source) {
                        int delay = customization.getC2SPushDelay();
                        Vector ss = new Vector();
                        ss.addElement(appSource);
                        hsc.synchronize("", ss, delay, true);
                        break;
                    }
                }
            }
        }
    }
}
 
