/*
 * Funambol is a mobile platform developed by Funambol, Inc. 
 * Copyright (C) 2003 - 2007 Funambol, Inc.
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License version 3 as published by
 * the Free Software Foundation with the addition of the following permission 
 * added to Section 15 as permitted in Section 7(a): FOR ANY PART OF THE COVERED
 * WORK IN WHICH THE COPYRIGHT IS OWNED BY FUNAMBOL, FUNAMBOL DISCLAIMS THE 
 * WARRANTY OF NON INFRINGEMENT  OF THIRD PARTY RIGHTS.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Affero General Public License 
 * along with this program; if not, see http://www.gnu.org/licenses or write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA.
 * 
 * You can contact Funambol, Inc. headquarters at 643 Bair Island Road, Suite 
 * 305, Redwood City, CA 94063, USA, or at email address info@funambol.com.
 * 
 * The interactive user interfaces in modified source and object code versions
 * of this program must display Appropriate Legal Notices, as required under
 * Section 5 of the GNU Affero General Public License version 3.
 * 
 * In accordance with Section 7(b) of the GNU Affero General Public License
 * version 3, these Appropriate Legal Notices must retain the display of the
 * "Powered by Funambol" logo. If the display of the logo is not reasonably 
 * feasible for technical reasons, the Appropriate Legal Notices must display
 * the words "Powered by Funambol".
 */

package com.funambol.blackberry.controller;

import com.funambol.blackberry.BBVersion;
import com.funambol.blackberry.Configuration;
import com.funambol.blackberry.Customization;
import com.funambol.blackberry.SyncClient;
import com.funambol.blackberry.SyncEngine;
import com.funambol.blackberry.ui.DisplayManager;
import com.funambol.util.ThreadPool;
import com.funambol.util.Log;
import com.funambol.util.StringUtil;
import com.funambol.util.AppProperties;
import com.funambol.util.StaticDataHelper;
import com.funambol.util.TransportAgent;
import com.funambol.updater.Updater;
import com.funambol.updater.BasicUpdaterConfig;
import com.funambol.updater.UpdaterListener;

import java.util.Calendar;
import net.rim.blackberry.api.browser.Browser;

import java.util.Date;
import net.rim.device.api.i18n.DateFormat;

/**
 * This class is the client interface to the automatic update library.
 */
public class UpdaterManager implements UpdaterListener {

    private static final String TAG_LOG = "UpdaterManager";

    /** Default timing definition */
    public static final long DEFAULT_CHECK_INTERVAL = Customization.getInstance().getCheckUpdtIntervalDefault();
    public static final long DEFAULT_REMINDER_INTERVAL = Customization.getInstance().getReminderUpdtIntervalDefault();

    /**
     * The BlackBerry Sync Client component name:
     *  - bbpim for BlackBerry OS from 4.2.1 to 4.6
     *  - bbpim_os47 for BlackBerry OS 4.7
     */
    private String componentName           = "bbpim";

    private Updater updater                = null;

    private Configuration configuration    = null;

    /** The update configuration */
    private UpdaterManagerConfig config    = null;

    /** Update procedure timing definition */
    private long checkInterval = DEFAULT_CHECK_INTERVAL;
    private long reminderInterval = DEFAULT_REMINDER_INTERVAL;

    /** The UpdaterManager singleton */
    private static UpdaterManager instance = null;

    private Controller controller;

    private boolean alertShown = false;

    private final int UPDATE_NOW_VALUE = 0;
    private final int UPDATE_LATER_VALUE = 1;
    private final int SKIP_UPDATE_VALUE = 2;

    private final String UPDATE_NOW_LABEL = StaticDataHelper.getLanguage("update_command-now");
    private final String UPDATE_LATER_LABEL = StaticDataHelper.getLanguage("update_command-later");
    private final String SKIP_UPDATE_LABEL = StaticDataHelper.getLanguage("update_command-skip");
    private final String EXPIRED_UPDATE_LABEL = StaticDataHelper.getLanguage("update_command-expired");

    private final String UPDATE_POPUP_MESSAGE = StaticDataHelper.getLanguage("message_update-available");
    private final String UPDATE_POPUP_MESSAGE_REQUIRED_1 = StaticDataHelper.getLanguage("message_update-required-1");
    private final String UPDATE_POPUP_MESSAGE_REQUIRED_2 = StaticDataHelper.getLanguage("message_update-required-2");
    private final String UPDATE_POPUP_MESSAGE_EXPIRED = StaticDataHelper.getLanguage("message_update-expired");

    private final String MAND_RECOM_POPUP = "mand_rec";
    private final String EXPIRED_POPUP = "expired";
    private final String OPTIONAL_POPUP = "optional";
    
    /**
     * Get the singleton of the <code>UpdaterManager</code>
     * @return The <code>UpdaterManager</code> singleton
     */
    public static UpdaterManager getInstance() {
        if (instance == null) {
            instance = new UpdaterManager();
        }
        return instance;
    }

    /**
     * Set the client <code>Controller</code> reference.
     * @param controller The <code>Controller</code> reference
     */
    public void setController(Controller controller){
        this.controller = controller;
    }

    /**
     * The <code>Thread</code> used to run the Update procedure
     */
    private class UpdaterThread extends Thread {
        
        public UpdaterThread() { }

        public void run() {
            updater.check();
        }
    }

    /**
     * Manages the Updater configuration parameters.
     */
    private class UpdaterManagerConfig extends BasicUpdaterConfig {

        /**
         * Load the Update manager configuration
         */
        public void load() {
            try {
                Configuration.load();
            } catch (Exception ce) {
                Log.error(TAG_LOG, "Loading updater config failed: ", ce);

                // We shall create a new config
                String server = Customization.getInstance().getServerUriDefault();
                server = StringUtil.extractAddressFromUrl(server);

                Log.debug(TAG_LOG, "Server url: " + server);

                super.setUrl(server);
                checkInterval = Customization.getInstance().getCheckUpdtIntervalDefault();
                setCheckInterval(checkInterval);
                reminderInterval = Customization.getInstance().getReminderUpdtIntervalDefault();
                setReminderInterval(reminderInterval);
            }
            String server = Configuration.getSyncUrl();
            server = StringUtil.extractAddressFromUrl(server);
            Log.debug(TAG_LOG, "Server url: " + server);
            setUrl(server);
            setDownloadUrl(Configuration.getDownloadUrl());
            setType(Configuration.getUpdateType());
            setAvailableVersion(Configuration.getAvailableVersion());
            setLastCheck(Configuration.getLastUpdateCheck());
            setCheckInterval(Configuration.getCheckInterval());
            setReminderInterval(Configuration.getReminderInterval());
            setLastReminder(Configuration.getLastReminder());
            setSkip(Configuration.getSkip());
            setActivationDate(configuration.getActivationDate());
        }

        /**
         * Save the Update manager configuration
         */
        public void save() {
            try {
                Configuration.setDownloadUrl(getDownloadUrl());
                Configuration.setUpdateType(getType());
                Configuration.setAvailableVersion(getAvailableVersion());
                Configuration.setLastUpdateCheck(getLastCheck());
                Configuration.setCheckInterval(getCheckInterval());
                Configuration.setReminderInterval(getReminderInterval());
                Configuration.setLastReminder(getLastReminder());
                Configuration.setSkip(getSkip());
                configuration.setActivationDate(getActivationDate());
                Configuration.saveConfiguration();
            } catch (Exception ce) {
                Log.error(TAG_LOG, "Saving updater config failed ", ce);
            }
        }
    }

    /**
     * Default constructor
     */
    private UpdaterManager() {
        String version = BBVersion.VERSION;
        configuration = Configuration.getInstance();
        config = new UpdaterManagerConfig();
        config.load();
        String [] osVersion = StaticDataHelper.getOSVersion();
        int os = Integer.parseInt(osVersion[0] + osVersion[1]);
        if(os > 46){
            componentName = componentName + "_os47";
        }
        updater = new Updater(config, version, componentName);
        String customizedReqUri = Customization.getInstance().getUpdaterRequestUri();
        if (customizedReqUri != null) {
            updater.setRequestUri(customizedReqUri);
        }
        updater.setListener(this);
    }

    /**
     * Check for updates
     */
    public void check() {
        check(true);
    }

    public void check(boolean spawnThread) {
        // The update check is executed in a separate thread
        UpdaterThread updThread = new UpdaterThread();
        updThread.setPriority(Thread.MIN_PRIORITY);
        if (spawnThread) {
            updThread.start();
        } else {
            updThread.run();
        }
    }

    /**
     * @see com.funambol.updater.UpdaterListener
     */
    public void optionalUpdateAvailable(String version) {
        promptUpdateAvailable(UPDATE_POPUP_MESSAGE, OPTIONAL_POPUP);
    }

    /**
     * @see com.funambol.updater.UpdaterListener
     */
    public void recommendedUpdateAvailable(String version) {
        promptUpdateAvailable(UPDATE_POPUP_MESSAGE, MAND_RECOM_POPUP);
    }

    /**
     * @see com.funambol.updater.UpdaterListener
     */
    public void mandatoryUpdateAvailable(String version) {
        Log.debug(TAG_LOG, "Mandatory update available");
        
        promptUpdateAvailable( 
                UPDATE_POPUP_MESSAGE_REQUIRED_1+" "+
                (DateFormat.getInstance(DateFormat.DATE_LONG)).formatLocal(config.getActivationDate())+
                UPDATE_POPUP_MESSAGE_REQUIRED_2, MAND_RECOM_POPUP);
    }


    public void promptExpiredUpdate() {
        Log.debug(TAG_LOG, "Mandatory update available");
        promptUpdateAvailable(UPDATE_POPUP_MESSAGE_EXPIRED,EXPIRED_POPUP);
    }


    /*
     * If a Mandatory update is expired needs to upgrade application
     */
    public boolean needsToUpdate(){
        long now = System.currentTimeMillis();
        if(StringUtil.equalsIgnoreCase(config.getType(),"mandatory") &&
                updater.isNewVersionAvailable() &&
                now > config.getActivationDate()){
            return true;
        }else{
            return false;
        }


    }

    public boolean dialogShown(){
        return alertShown;
    }


    /**
     * Display an update prompt.
     *
     * @param version
     *            The current software version
     * @param message
     *            The message to dsiplay
     * @param allowSkip
     *            If skipping the update is an option
     */
    private void promptUpdateAvailable(String message, String popupType) {
        // In this client an update is available only if we have a new
        // version and a valid url.
        
        controller.toForeground();
        if (alertShown) {
            Log.info(TAG_LOG, "An update alert is already on screen, do not add another one");
            return;
        }

        String url = getDownloadUrl();
        if (url != null && url.length() > 0) {
            DialogController dialControll = controller.getDialogController();
            String [] labels;
            int [] values;
            if (popupType.equals(OPTIONAL_POPUP)) {
                labels = new String[]{UPDATE_NOW_LABEL,UPDATE_LATER_LABEL, SKIP_UPDATE_LABEL};
                values = new int[]{UPDATE_NOW_VALUE,UPDATE_LATER_VALUE,SKIP_UPDATE_VALUE};
            } else if(popupType.equals(MAND_RECOM_POPUP)) {
                labels = new String[]{UPDATE_NOW_LABEL,UPDATE_LATER_LABEL};
                values = new int[]{UPDATE_NOW_VALUE,UPDATE_LATER_VALUE};
            } else {
                //expired update
                labels = new String[]{EXPIRED_UPDATE_LABEL};
                values = new int[]{UPDATE_NOW_VALUE};
            }
            alertShown = true;
            try {
                
                int selection = dialControll.promptSelection(message,
                        labels,
                        values,
                        UPDATE_NOW_VALUE,
                        true,
                        DisplayManager.NO_LIMIT);

                if (selection == UPDATE_NOW_VALUE){
                    this.userSelectedNow();
                }else if (selection == UPDATE_LATER_VALUE){
                    this.userSelectedLater();
                }else{
                    //selected SKIP_UPDATE_VALUE
                    this.userSelectedSkip();
                }
            } finally {
                alertShown = false;
            }
        }
    }

    /**
     * Update the last update reminder timestamp.
     */
    private void updateLastReminder() {
        long now = System.currentTimeMillis();
        Log.info(TAG_LOG, "A reminder will be showed after: " +
                 new Date(now+ config.getReminderInterval()));
        updater.setLastReminder(now);
        config.save();
    }

    /**
     * Tells the <code>UpdaterManager</code> that the user choosed "Now" to
     * the update message.
     */
    public void userSelectedNow() {
        Log.info(TAG_LOG, "Update now");
        this.update();
    }

    /**
     * Tells the <code>UpdaterManager</code> that the user choosed "Later" to
     * the update message.
     */
    public void userSelectedLater() {
        Log.info(TAG_LOG, "Update later");
        this.updateLastReminder();
    }

    /**
     * Tells the <code>UpdaterManager</code> that the user choosed "Skip" to
     * the update message.
     */
    public void userSelectedSkip() {
        Log.info(TAG_LOG, "Skip update");
        updater.setSkip();
    }

    /**
     * Check wheter an update is currently available.
     * @return
     */
    public boolean isUpdateAvailable() {
        // In this client an update is available only if we have a new
        // version and a valid url.
        String url = getDownloadUrl();
        if (url != null && url.length() > 0) {
            return updater.isUpdateAvailable();
        } else {
            return false;
        }
    }

    /**
     * Get the url where the new client version is located.
     * @return The update download url.
     */
    public String getDownloadUrl() {
        return config.getDownloadUrl();
    }

    /**
     * Set the url where the new client version is located.
     * @param url The update download url.
     */
    public void setUrl(String url) {
        if(!url.equals(config.getUrl())) {
            config.setUrl(url);
            // Since a new url has been set, we also reset our timers to force a new
            // check and possibly a new reminder
            config.setLastReminder(0);
            config.setLastCheck(0);
            config.save();
        }
    }

    /**
     * Update the client. It Opens the default Browser to the update download
     * location.
     */
    public void update(){
        Log.info(TAG_LOG, "Updating client from: " + config.getDownloadUrl());
        Browser.getDefaultSession().displayPage(config.getDownloadUrl());
    }

    /*
     * Reset all Update parameters
     */
    public void resetUpdateParams(){
        config.setDownloadUrl("");
        config.setType("");
        config.setAvailableVersion("");
        config.setLastCheck(0);
        config.setCheckInterval(Configuration.getCheckInterval());
        config.setLastReminder(0);
        config.setSkip(false);
        config.setActivationDate(0);
        config.save();
    }

}