/*
 * Funambol is a mobile platform developed by Funambol, Inc.
 * Copyright (C) 2003 - 2007 Funambol, Inc.
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License version 3 as published by
 * the Free Software Foundation with the addition of the following permission
 * added to Section 15 as permitted in Section 7(a): FOR ANY PART OF THE COVERED
 * WORK IN WHICH THE COPYRIGHT IS OWNED BY FUNAMBOL, FUNAMBOL DISCLAIMS THE
 * WARRANTY OF NON INFRINGEMENT  OF THIRD PARTY RIGHTS.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program; if not, see http://www.gnu.org/licenses or write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA.
 * 
 * You can contact Funambol, Inc. headquarters at 643 Bair Island Road, Suite
 * 305, Redwood City, CA 94063, USA, or at email address info@funambol.com.
 *
 * The interactive user interfaces in modified source and object code versions
 * of this program must display Appropriate Legal Notices, as required under
 * Section 5 of the GNU Affero General Public License version 3.
 *
 * In accordance with Section 7(b) of the GNU Affero General Public License
 * version 3, these Appropriate Legal Notices must retain the display of the
 * "Powered by Funambol" logo. If the display of the logo is not reasonably
 * feasible for technical reasons, the Appropriate Legal Notices must display
 * the words "Powered by Funambol".
 */

package com.funambol.blackberry.helper;

import javax.microedition.pim.PIMItem;
import javax.microedition.pim.PIMList;
import javax.microedition.pim.PIMException;
import javax.microedition.pim.Event;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.OutputStream;
import java.util.Calendar;

import net.rim.blackberry.api.pdap.BlackBerryEvent;

import com.funambol.blackberry.source.AdditionalItemsAttributes;
import com.funambol.blackberry.source.CalendarAdditionalAttributes;
import com.funambol.common.pim.icalendar.ICalendarParserListener;
import com.funambol.common.pim.icalendar.ICalendarFormatter;
import com.funambol.common.pim.icalendar.ICalendarSyntaxParser;
import com.funambol.common.pim.icalendar.ICalendar;
import com.funambol.common.pim.PimUtils;

import com.funambol.util.Log;

/**
 * This class extends the EventHelper to help on iCalendar items handling.
 * It redefines the iCalendar parser & formatter in order to save additional
 * fields not directly supported by the basic JSR75 implementation. These fields
 * are saved into the BlackBerry PIMItem object if possible, otherwise into an
 * AdditionalItemsAttributes object.
 */
public class ICalEventHelper extends EventHelper {
    
    private static final int[] SUPPORTED_FIELDS = {
        BlackBerryEvent.START,
        BlackBerryEvent.END,
        BlackBerryEvent.ALLDAY,
        BlackBerryEvent.SUMMARY,
        BlackBerryEvent.LOCATION,
        BlackBerryEvent.NOTE,
        BlackBerryEvent.ALARM,
        BlackBerryEvent.CLASS,
        BlackBerryEvent.REVISION,
        BlackBerryEvent.ATTENDEES
    };

    private CalendarAdditionalAttributes additionalAttributes = null;

    /**
     * This class extends the default CalendarParserListener in order to save
     * additional fields not directly supported by the basic JSR75
     * implementation.
     */
    private class BBEventParserListener extends ICalendarParserListener {

        public BBEventParserListener(Event event) {
            super(event);
        }

        protected void setTZID(String value) {
            additionalAttributes.setTZID(value);
        }

        protected void setTZOffset(long offset) {
            additionalAttributes.setTZOffset(offset);
        }

        protected void setAllDay(boolean allday) {
            if(isSupportedField(BlackBerryEvent.ALLDAY)) {
                Log.trace("[BBEventParserListener.setAllDay] Value: " + allday);
                pimItem.addBoolean(BlackBerryEvent.ALLDAY, BlackBerryEvent.ATTR_NONE, allday);
            }
        }

        protected void addAttendee(String value) {
            if(isSupportedField(BlackBerryEvent.ATTENDEES)) {
                Log.trace("[BBEventParserListener.addAttendee]");
                // Cut the mail address uri "MAILTO:"
                int index = value.indexOf(ICalendar.ATTENDEE_EMAIL_URI + ":");
                if(index >= 0) {
                    value = value.substring(index + ICalendar.ATTENDEE_EMAIL_URI.length() + 1);
                    pimItem.addString(BlackBerryEvent.ATTENDEES,
                        BlackBerryEvent.ATTR_NONE,
                        value);
                    Log.debug("[BBEventParserListener.addAttendee] Added mail address: " + value);
                }
                else {
                    Log.error("[BBEventParserListener.addAttendee] Failed to add mail address: " + value);
                }
            }
        }
    }

    /**
     * This class extends the default ICalendarFormatter in order to format
     * additional fields not directly supported by the basic JSR75
     * implementation.
     */
    private class BBEventFormatter extends ICalendarFormatter {

        private CalendarAdditionalAttributes additionals;

        public BBEventFormatter(String defaultCharset, CalendarAdditionalAttributes add) {
            super(defaultCharset);
            additionals = add;
        }
        protected String getTZID(PIMItem pimItem) {
            String tzid = additionals.getTZID();
            Log.trace("[BBEventFormatter.getTZID] Value: " + tzid);
            return tzid;
        }
        protected boolean isAllDay(PIMItem pimItem) {
            if(pimItem instanceof Event) {
                if(isSupportedField(BlackBerryEvent.ALLDAY) &&
                   pimItem.countValues(BlackBerryEvent.ALLDAY) > 0) {
                    boolean allday = pimItem.getBoolean(BlackBerryEvent.ALLDAY, 0);
                    Log.trace("[BBEventFormatter.isAllDay] Value: " + allday);
                    return allday;
                }
            }
            return false;
        }
        protected long getTZOffset(PIMItem pimItem) {
            long tzoffset = additionals.getTZOffset();
            Log.trace("[BBEventFormatter.getTZOffset] Value: " + tzoffset);
            return tzoffset;
        }
        protected void formatAttendees(PIMItem pimItem, OutputStream os) throws PIMException {
            Log.trace("[BBEventFormatter.formatAttendees]");
            if (isSupportedField(BlackBerryEvent.ATTENDEES)) {
                String email = null;
                for(int i=0; i<pimItem.countValues(BlackBerryEvent.ATTENDEES); i++) {
                    email = pimItem.getString(BlackBerryEvent.ATTENDEES, i);
                    Log.trace("[BBEventFormatter.formisallatAttendees] Format email: " + email);
                    PimUtils.println(os, ICalendar.ATTENDEE + ":" + 
                                     ICalendar.ATTENDEE_EMAIL_URI + ":" + email);
                }
            }
        }

        /**
         * The end date shouldn't be shifted becouse it is already set to one
         * day after the end date (see BlackBerryEvent.ALLDAY documentaion)
         * @param endDate
         * @return
         */
        protected long fixEndDate(long endDate) {
            return endDate;
        }
    }

    /**
     * Constructor
     */
    public ICalEventHelper(PIMList list, EventMailListener listener, int pastlimit, int futurelimit) {
        super(list, listener, pastlimit, futurelimit);
    }

    /*
     * (non-Javadoc)
     * @see com.funambol.blackberry.helper.PIMItemHelper#itemToString(PIMItem item, AdditionalItemsAttributes additionals)
     */
    public String itemToString(PIMItem item, AdditionalItemsAttributes additionals) throws PIMException {
        if (item instanceof Event) {
            ByteArrayOutputStream ostream = new ByteArrayOutputStream();
            BBEventFormatter formatter =  new BBEventFormatter(ICalendar.UTF8, (CalendarAdditionalAttributes)additionals);
            formatter.format(item, ostream, true);
            return ostream.toString();
        }
        return null;
    }

    /*
     * (non-Javadoc)
     * @see com.funambol.blackberry.helper.PIMItemHelper#fillItem(PIMItem item, String pimString, AdditionalItemsAttributes additionals)
     */
    public AdditionalItemsAttributes fillItem(PIMItem item, String pimString, AdditionalItemsAttributes additionals) throws PIMException {
        
        Log.trace("[ICalEventHelper.fillItem]");
        if(additionals == null){
            Log.debug("[ICalEventHelper.fillItem] Additionals null, creating new one");
            additionalAttributes = new CalendarAdditionalAttributes();
        } else {
            additionalAttributes = (CalendarAdditionalAttributes) additionals;
        }
        if (item instanceof Event) {
            ByteArrayInputStream is = new ByteArrayInputStream(pimString.getBytes());
            try {
                Event newEvent = (Event)item;
                BBEventParserListener listener = new BBEventParserListener(newEvent);
                ICalendarSyntaxParser parser = new ICalendarSyntaxParser(is);
                parser.setListener(listener);
                parser.parse();
            } catch (Exception pe) {
                Log.debug("[ICalEventHelper.fillItem] Exception while parsing the item " + pe.toString());
                throw new PIMException(pe.toString());
            }
        }
        return additionalAttributes;
    }
    
    /*
     * (non-Javadoc)
     * @see com.funambol.blackberry.helper.PIMItemHelper#clearPIMItem(PIMItem _item)
     */
    public void clearItem(PIMItem _item) throws PIMException {
        PIMItemHelper.clearPIMItem(_item, BlackBerryEvent.UID, SUPPORTED_FIELDS);
        ((Event)_item).setRepeat(null);
    }

    /*
     * (non-Javadoc)
     * @see com.funambol.blackberry.helper.PIMItemHelper#copyUnsupportedFields(PIMItem origItem, PIMItem item)
     */
    public void copyUnsupportedFields(PIMItem origItem, PIMItem item) throws PIMException {
        copyUnsupportedFields(origItem, item, BlackBerryEvent.UID, SUPPORTED_FIELDS);
    }

    /*
     * (non-Javadoc)
     * @see com.funambol.blackberry.helper.PIMItemHelper#isInSyncRange(PIMItem _item)
     */
    public boolean isInSyncRange(PIMItem _item) throws PIMException {

        long now = System.currentTimeMillis();

        Calendar start = getFieldDate(_item, BlackBerryEvent.START, false, false, false);
        Calendar end = getFieldDate(_item, BlackBerryEvent.END, true, false, false);
 
        if (start!=null && end!=null) {
            if (end.getTime().getTime() < (now-limitPast*EventHelper.MONTH_FACTOR)
                || start.getTime().getTime() > (now+limitFuture*EventHelper.MONTH_FACTOR))
            {
                return false;
            }
        }
        return true;
    }

    public AdditionalItemsAttributes createAdditionals(PIMItem newItem){
        Log.trace("[ICalEventHelper.createAdditionals]");
        if (newItem instanceof Event) {
            additionalAttributes = new CalendarAdditionalAttributes();
            return additionalAttributes;
        }
        return null;
    }
}
    

   
