/*
 * Funambol is a mobile platform developed by Funambol, Inc.
 * Copyright (C) 2010 Funambol, Inc.
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License version 3 as published by
 * the Free Software Foundation with the addition of the following permission
 * added to Section 15 as permitted in Section 7(a): FOR ANY PART OF THE COVERED
 * WORK IN WHICH THE COPYRIGHT IS OWNED BY FUNAMBOL, FUNAMBOL DISCLAIMS THE
 * WARRANTY OF NON INFRINGEMENT  OF THIRD PARTY RIGHTS.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program; if not, see http://www.gnu.org/licenses or write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA.
 * 
 * You can contact Funambol, Inc. headquarters at 643 Bair Island Road, Suite
 * 305, Redwood City, CA 94063, USA, or at email address info@funambol.com.
 *
 * The interactive user interfaces in modified source and object code versions
 * of this program must display Appropriate Legal Notices, as required under
 * Section 5 of the GNU Affero General Public License version 3.
 *
 * In accordance with Section 7(b) of the GNU Affero General Public License
 * version 3, these Appropriate Legal Notices must retain the display of the
 * "Powered by Funambol" logo. If the display of the logo is not reasonably
 * feasible for technical reasons, the Appropriate Legal Notices must display
 * the words "Powered by Funambol".
 */

package com.funambol.blackberry.source;

import java.util.Vector;
import java.util.Enumeration;
import java.io.IOException;

import com.funambol.platform.FileAdapter;
import com.funambol.storage.StringKeyValueFileStore;
import com.funambol.syncml.spds.SourceConfig;
import com.funambol.syncml.spds.SyncException;
import com.funambol.syncml.protocol.SyncMLStatus;
import com.funambol.syncml.client.ChangesTracker;
import com.funambol.syncml.client.CacheTracker;
import com.funambol.blackberry.Configuration;
import com.funambol.blackberry.controller.SynchronizationController;
import com.funambol.syncml.client.HttpUploader;
import com.funambol.syncml.client.TwoPhasesFileSyncSource;
import com.funambol.syncml.spds.SyncConfig;
import com.funambol.syncml.spds.SyncListener;
import com.funambol.util.Log;
import com.funambol.util.MD5;
import com.funambol.util.Base64;

/**
 * Extends TwoPhasesFileSyncSource in order to adopt the TwoPhases
 * synchronization in Pictures sync.
 */
public class TwoPhasesPhotoSyncSource extends TwoPhasesFileSyncSource
        implements ExtendedSyncSource {
    
    private static final String TAG_LOG = "TwoPhasesPhotoSyncSource";

    protected SynchronizationController syncController;
    protected boolean quotaReached;

    public TwoPhasesPhotoSyncSource(SourceConfig config, String extensions[],
            SyncConfig sConfig, String uploadUrl) {
        super(config, createTracker(), Configuration.getPicturesDir(), sConfig, uploadUrl);
        setSupportedExtensions(extensions); 
    }

    public void setController(SynchronizationController controller) {
        this.syncController = controller;
    }

    public void beginSync(int syncMode) throws SyncException {
        super.beginSync(syncMode);
        quotaReached = false;
    }

    public void reapplyConfiguration() {
        Log.trace(TAG_LOG, "reapplyConfiguration");
        this.syncConfig = Configuration.getSyncConfig();
        if (this.directory == null || !this.directory.equals(Configuration.getPicturesDir())) {
            this.tracker = createTracker();
            this.tracker.setSyncSource(this);
            this.directory = Configuration.getPicturesDir();
        }
    }

    public void clearTracker() throws IOException {
        Log.info(TAG_LOG, "Clearing photo tracker");
        String cacheName = getCacheName();
        // Delete the file
        FileAdapter fa = new FileAdapter(cacheName);
        if (fa.exists()) {
            fa.delete();
            fa.close();
        }
        // Sets a new tracker
        setTracker(createTracker());
    }

    public void setItemStatus(String key, int status) throws SyncException {

        if (status == SyncMLStatus.DEVICE_FULL) {
            // The user reached his quota on the server. Do not interrupt the
            // sync here because we need to complete the sync first (and upload
            // all pictures that received a positive status)
            quotaReached = true;
            // The user reached his quota on the server
            Log.info(TAG_LOG, "Server is full");
        }
        super.setItemStatus(key, status);
    }

    protected Enumeration getAllItemsKeys() throws SyncException {
        Log.trace(TAG_LOG, "getAllItemsKeys");
        // Scan the briefcase directory and return all keys
        try {
            FileAdapter dir = new FileAdapter(directory);
            Enumeration files = dir.list(false);
            Vector filteredList = new Vector();
            // Filter out the unwanted extensions
            while(files.hasMoreElements()) {
                String path = (String)files.nextElement();
                if (filterFile(path)) {
                    filteredList.addElement(directory + path);
                }
            }
            dir.close();
            return filteredList.elements();
        }
        catch (Exception e) {
            throw new SyncException(SyncException.CLIENT_ERROR, e.toString());
        }
    }

    private static String getCacheName() {
        String directory = Configuration.getPicturesDir();
        String sep = "";
        if (!directory.endsWith("/")) {
            sep = "/";
        }

        // The cache file name depends on the username and server, so that
        // we allow the user to switch account without losing track of what was
        // synchronized
        // TODO: shall we remove the old cache?
        // TODO: shall we hide the cache?
        String username = Configuration.getUsername();
        String syncUrl  = Configuration.getSyncUrl();

        byte cred[] = (username + ":" + syncUrl).getBytes();
        MD5 md5 = new MD5();
        byte[] hash = md5.calculateMD5(cred);
        String id = new String(Base64.encode(hash));

        // Replace the dir separator as this may cause issues
        id = id.replace('/', '_');
        String cacheName = directory + sep + "ps_" + id + ".txt";

        return cacheName;
    }

    private static ChangesTracker createTracker() {
        Log.trace(TAG_LOG, "createTracker");
        String cacheStoreFile = getCacheName();
        String directory = Configuration.getPicturesDir();
        StringKeyValueFileStore cacheStore = new StringKeyValueFileStore(cacheStoreFile);
        CacheTracker tracker = new PhotoTracker(cacheStore, directory);
        return tracker;
    }

    /**
     * Creates a specific uploader for BB devices.
     * 
     * @param config
     * @param uploadUrl
     * @param sourceUri
     * @param listener
     * @return
     */
    protected HttpUploader createUploader(SyncConfig config, String uploadUrl,
                                          String sourceUri, SyncListener listener)
    {
        BBHttpUploader uploader = new BBHttpUploader(Configuration.getSyncConfig(), uploadUrl, sourceUri, listener);
        return uploader;
    }

    protected void uploadItems() throws SyncException {
        super.uploadItems();
        if (quotaReached) {
            // Generate the exception to signal the condition
            throw new SyncException(SyncException.DEVICE_FULL, "Server is full");
        }
    }

    /**
     * Implements a specific HttpUploader used for BB devices in order to
     * catch the sync cancelled event.
     */
    private class BBHttpUploader extends HttpUploader {

        public BBHttpUploader(SyncConfig syncConfig, String uploadUrl,
                String remoteUri, SyncListener listener) {
            super(syncConfig, uploadUrl, remoteUri, listener);
        }

        protected boolean uploadCancelled() {
            return (syncController != null && syncController.isCancelled());
        }
    }
}
    
 
