/*
 * Funambol is a mobile platform developed by Funambol, Inc.
 * Copyright (C) 2009 Funambol, Inc.
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License version 3 as published by
 * the Free Software Foundation with the addition of the following permission
 * added to Section 15 as permitted in Section 7(a): FOR ANY PART OF THE COVERED
 * WORK IN WHICH THE COPYRIGHT IS OWNED BY FUNAMBOL, FUNAMBOL DISCLAIMS THE
 * WARRANTY OF NON INFRINGEMENT  OF THIRD PARTY RIGHTS.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program; if not, see http://www.gnu.org/licenses or write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA.
 * 
 * You can contact Funambol, Inc. headquarters at 643 Bair Island Road, Suite
 * 305, Redwood City, CA 94063, USA, or at email address info@funambol.com.
 *
 * The interactive user interfaces in modified source and object code versions
 * of this program must display Appropriate Legal Notices, as required under
 * Section 5 of the GNU Affero General Public License version 3.
 *
 * In accordance with Section 7(b) of the GNU Affero General Public License
 * version 3, these Appropriate Legal Notices must retain the display of the
 * "Powered by Funambol" logo. If the display of the logo is not reasonably
 * feasible for technical reasons, the Appropriate Legal Notices must display
 * the words "Powered by Funambol".
 */

package com.funambol.blackberry.ui.rim;

import java.util.Vector;
import java.util.Enumeration;
import java.util.Stack;
import javax.microedition.io.Connector;
import javax.microedition.io.file.FileConnection;
import javax.microedition.io.file.FileSystemRegistry;

import net.rim.device.api.system.Bitmap;
import net.rim.device.api.system.KeyListener;
import net.rim.device.api.system.TrackwheelListener;
import net.rim.device.api.ui.Color;
import net.rim.device.api.ui.DrawStyle;
import net.rim.device.api.ui.Field;
import net.rim.device.api.ui.Font;
import net.rim.device.api.ui.Graphics;
import net.rim.device.api.ui.Keypad;
import net.rim.device.api.ui.MenuItem;
import net.rim.device.api.ui.component.LabelField;
import net.rim.device.api.ui.component.TextField;
import net.rim.device.api.ui.component.ListField;
import net.rim.device.api.ui.component.ListFieldCallback;
import net.rim.device.api.ui.component.Menu;
import net.rim.device.api.ui.container.MainScreen;

import com.funambol.blackberry.ui.Screen;
import com.funambol.blackberry.BBVersion;
import com.funambol.blackberry.Configuration;
import com.funambol.blackberry.Customization;
import com.funambol.blackberry.Constants;
import com.funambol.blackberry.SyncClient;
import com.funambol.blackberry.AppSyncSourceManager;
import com.funambol.blackberry.AppSyncSource;
import com.funambol.blackberry.ExternalAppManager;
import com.funambol.blackberry.controller.Controller;
import com.funambol.blackberry.ui.HomeScreen;
import com.funambol.util.Log;
import com.funambol.util.StaticDataHelper;


/**
 * This class implements a file browser screen. Such a screen allows
 * users to navigate the file system and choose files or directories.
 * At the moment it supports single selection, but it can be extended if needed.
 * The screen does not rely on a controller, as its logic its fairly simple. It
 * behaves like one of the standard components available on many platform that
 * do the same work.
 * The behavior of this file browser resembles as much as possible the RIM
 * native file browser. The screen displays a list of directories and files,
 * separated by a separator line. The dialog can be configured to filter out
 * files or directories.
 * A menu allows the user to move "up", accept the selected item or cancel. The
 * "up" operation can also be performed via the "back" key.
 */
public class RimFileBrowser extends MainScreen implements Screen {

    public static final int      DIR_MODE       = 0;
    public static final int      FILE_MODE      = 1;
    public static final int      ALL_MODE       = 2;

    public static final int      SELECT_DIR_MODE  = 0;
    public static final int      SELECT_FILE_MODE = 1;

    private static final String  ROOT           = "/";

    protected FileBrowserList    listField      = null;

    private TextField            titleField     = null;

    private String               currDirName    = ROOT;

    private int                  mode           = ALL_MODE;
    private int                  selectMode     = SELECT_FILE_MODE;

    private Vector               extensions     = null;

    private FileBrowserMenu      menu           = new FileBrowserMenu();

    private boolean              quit           = false;

    private Stack                dirStack       = new Stack();

    private RimFileBrowserListener listener     = null;

    private Bitmap               folderIcon     = null;

    private Controller           controller     = null;

    private boolean              caseSensitive  = true;

    /**
     * Create a FileBrowser in ALL MODE in the root directory.
     */
    public RimFileBrowser(Controller controller) {
        this.controller = controller;
        initialize();
    }

    /**
     * Create a FileBrowser the root directory.
     * @param mode can be DIR_MODE (directories only), FILE_MODE (files only)
     *             ALL_MODE (directories and files)
     */
    public RimFileBrowser(Controller controller, int mode) {
        this(controller);
        if (mode >= DIR_MODE && mode <= FILE_MODE) {
            this.mode = mode;
        }
    }

    /**
     * Create a FileBrowser the root directory.
     * @param mode can be DIR_MODE (directories only), FILE_MODE (files only)
     *             ALL_MODE (directories and files)
     * @param selectMode can be SELECT_FILE_MODE to allow file selection
     *                   or SELECT_DIR_MODE to allow dir selection
     */
    public RimFileBrowser(Controller controller, int mode, int selectMode) {
        this(controller, mode);
        if (selectMode >= SELECT_DIR_MODE && selectMode <= SELECT_FILE_MODE) {
            this.selectMode = selectMode;
        }
    }


    /**
     * Create a FileBrowser the root directory.
     * @param mode can be DIR_MODE (directories only), FILE_MODE (files only)
     *             ALL_MODE (directories and files)
     * @param selectMode can be SELECT_FILE_MODE to allow file selection
     *               or SELECT_DIR_MODE to allow dir selection
     * @param extensions are the file extensions that are displayed (all other
     *                   files are discarded)
     */
    public RimFileBrowser(Controller controller, int mode, int selectMode, Vector extensions) {
        this(controller, mode, selectMode);
        this.extensions = extensions;
    }

    public void setCaseSensitive(boolean caseSensitive) {
        this.caseSensitive = caseSensitive;
    }

    /**
     * This method is invoked by the system to create the screen menu
     */ 
    public void makeMenu(Menu theMenu, int instance) {
        String entry = null;
        if (listField != null) {
            entry = listField.getSelectedEntry();
        }
        menu.makeMenu(theMenu, entry);
    }

    /**
     * Set a listener for this screen. The listener is notified on
     * events that need to be processed outside of this class.
     * @see RimFileBrowserListener.
     *
     * @param listener the listener (null to void it)
     */ 
    public void setListener(RimFileBrowserListener listener) {
        this.listener = listener;
    }

    /**
     * Set the current path. The file browser moves to the given path and
     * the existing navigation history is discarded.
     *
     * @param path the new path. This path must be a valid path, without the
     *             protocol prefix. For example it can be: /store/home/
     */
    public void setPath(String path) {
        // Just perform a minimal sanity check and revert to ROOT if path
        // is invalid
        try {
            FileConnection fileConn = (FileConnection)Connector.open("file://" + path, Connector.READ);
            fileConn.close();
        } catch (Exception e) {
            path = ROOT;
        }
        currDirName = path;
        listField.changeDirectory(0);
        // We must build the DirInfo stack otherwise we won't be able to
        // navigate backward
        dirStack = new Stack();
        int idx = path.indexOf("/");
        String partialPath = ROOT;
        if (idx != -1) {
            DirInfo dirInfo = new DirInfo(partialPath);
            dirStack.push(dirInfo);
        }

        while (idx != -1) {
            int nextIdx = path.indexOf("/", idx + 1);

            if (nextIdx != -1 && nextIdx != (path.length() - 1)) {
                String entry;
                entry = path.substring(idx + 1, nextIdx + 1);
                partialPath = partialPath + entry;
                DirInfo dirInfo = new DirInfo(partialPath);
                dirStack.push(dirInfo);
            }
            idx = nextIdx;
        }
    }

    /**
     * This method is invoked by either the system or the application. It is
     * meant to handle the closing event. When the "back" button is pressed, the
     * system invokes this method. If the browser is on the root directory, then
     * the screen is terminated. Otherwise it goes "up" one level.
     */
    public boolean onClose() {
        if (!quit && !ROOT.equals(currDirName)) {
            // Change to the previous directory
            up();
            return false; 
        }
        return quit();
    }

    protected void onDisplay() {

        Log.trace("OnDisplay");
        int rowHeight = getFont().getHeight() + 4;
        listField.setRowHeight(rowHeight);
        super.onDisplay();
    }

    private void buttonPressed(int index) {
        Log.trace("[RimFileBrowser.buttonPressed] " + index);

        // Update the stack of directories
        int selectedIndex = listField.getSelectedIndex();
        String entryName = listField.getSelectedEntry();
        if (entryName != null && isDirectory(entryName)) {
            DirInfo dirInfo = new DirInfo(currDirName);
            dirInfo.setSelected(selectedIndex);
            dirStack.push(dirInfo);

            // If this a directory then we change into it
            currDirName = currDirName + entryName;
            listField.changeDirectory(0);
        }
    }

    protected boolean navigationClick(int status, int time) {
        Log.trace("[RimFileBrowser.navigationClick]");
        if (status == 0) {
            buttonPressed(listField.getSelectedIndex());
            return true;
        }
        return false;
    }

    protected void choose() {
        // The user selected the current entry and is about to quit the screen
        // If there is a listener notify it, and check if the selection is
        // acceptable
        boolean accepted = true;
        if (listener != null) {
            Vector selectedEntries = new Vector();
            selectedEntries.addElement(currDirName);
            accepted = listener.chosen(selectedEntries);
        }

        if (accepted) {
            quit = true;
            onClose();
        }
    }

    protected boolean quit() {
        // The user selected the current entry and is about to quit the screen
        // If there is a listener notify it, and check if the selection is
        // acceptable
        boolean accepted = true;
        // This is cancel event unless we passed through the choose method which
        // sets the quit flag
        if (!quit && listener != null) {
            accepted = listener.cancelled();
        }

        if (accepted) {
            quit = true;
            try {
                controller.getScreenDisplayController().hideScreen(this);
                return true;
            } catch (Exception e) {
                Log.error("Cannot close FileBrowser " + e.toString());
                return false;
            }
        } else {
            return false;
        }
    }

    protected void up() {
        DirInfo dirInfo = (DirInfo)dirStack.pop();
        if (dirInfo == null) {
            // We are already at the top level, we should quit
            // and the code should not get here. But just to be
            // conservative we ignore this case
            return;
        }
        currDirName = dirInfo.getPath();
        int previousSelection = dirInfo.getSelected();
        if (previousSelection == -1) {
            listField.changeDirectory(0);
        } else {
            listField.changeDirectory(dirInfo.getSelected());
        }
    }

    private void resetTitle() {
        String title = StaticDataHelper.getLanguage("select");
        if (!ROOT.equals(currDirName)) {
            title = title + "\r\n" + currDirName;
        }
        titleField.setText(title);
    }

    private void initialize() {

        Log.trace("[RimFileBrowser.initialize]");
        
        this.setFont(RimHelper.getDefaultFont());

        // Pre load the icons
        folderIcon = Bitmap.getBitmapResource(Customization.getInstance().getFolderIcon());
        
        // The title depends on the current directory
        titleField = new TextField(TextField.NON_FOCUSABLE);
        titleField.setEditable(false);
        titleField.setFont(getFont().derive(StaticDataHelper.TITLE_FONT_STYLE,
                StaticDataHelper.TITLE_FONT_SIZE));
        resetTitle();
        setTitle(titleField);

        listField = new FileBrowserList();
        add(listField);
    }





    private class FileBrowserListener implements KeyListener, TrackwheelListener {

        private final static int KEYCODE_ENTER = Keypad.KEY_ENTER;

        // KeypadListener
        public boolean keyChar(char key, int status, int time) {
            return false;
        }

        public boolean keyDown(int keycode, int time) {
            int mapped = Keypad.map(keycode);
            if (mapped == KEYCODE_ENTER || keycode == KEYCODE_ENTER) {
                buttonPressed(listField.getSelectedIndex());
                return true;
            }
            return false;
        }

        public boolean keyRepeat(int keycode, int time) {
            return false;
        }

        public boolean keyStatus(int keycode, int time) {
            return false;
        }

        public boolean keyUp(int keycode, int time) {
            return false;
        }

        public boolean trackwheelClick(int status, int time) {
            buttonPressed(listField.getSelectedIndex());
            return true;
        }

        public boolean trackwheelRoll(int amount, int status, int time) {
            return false;
        }

        public boolean trackwheelUnclick(int status, int time) {
            return false;
        }
    }

    private boolean isDirectory(String entry) {
        return entry.endsWith("/");
    }

    private Enumeration getFileEntries() {

        Log.trace("[RimFileBrowser.getFileEntries]");

        Vector files = new Vector();
        Enumeration entries = null;

        if (mode != DIR_MODE) {

            try {
                int i = 0;
                if (ROOT.equals(currDirName)) {
                    entries = FileSystemRegistry.listRoots();
                } else {
                    FileConnection fileConn = null;
                    fileConn = (FileConnection)Connector.open("file://" + currDirName, Connector.READ);
                    entries = fileConn.list();
                    fileConn.close();
                }
            } catch (Exception e) {
                Log.error("Cannot get file list " + e.toString());
            }

            // Filter out directories
            while(entries != null && entries.hasMoreElements()) {
                String entry = (String)entries.nextElement();
                if (!isDirectory(entry)) {
                    // Check if this file extensions is accepted
                    if (extensions.size() > 0) {
                        for(int i=0;i<extensions.size();++i) {
                            String ext = (String)extensions.elementAt(i);

                            String entryComp;
                            if (!caseSensitive) {
                                ext = ext.toLowerCase();
                                entryComp = entry.toLowerCase();
                            } else {
                                entryComp = entry;
                            }

                            if (entryComp.endsWith(ext)) {
                                files.addElement(entry);
                                break;
                            }
                        }
                    } else {
                        files.addElement(entry);
                    }
                }
            }
        }
        entries = files.elements();

        return entries;
    }

    private Enumeration getDirectoryEntries() {

        Log.trace("[RimFileBrowser.getDirectoryEntries]");

        Enumeration entries = null;
        Vector dirs = new Vector();

        if (mode != FILE_MODE) {
            try {
                int i = 0;
                if (ROOT.equals(currDirName)) {
                    entries = FileSystemRegistry.listRoots();
                } else {
                    FileConnection fileConn = null;
                    fileConn = (FileConnection)Connector.open("file://" + currDirName, Connector.READ);
                    entries = fileConn.list();
                    fileConn.close();
                }
            } catch (Exception e) {
                Log.error("Cannot get file list " + e.toString());
            }

            // Filter out directories
            while(entries != null && entries.hasMoreElements()) {
                String entry = (String)entries.nextElement();
                if (isDirectory(entry)) {
                    dirs.addElement(entry);
                }
            }
        }
        entries = dirs.elements();

        return entries;
    }

    private class FileBrowserList extends ListField implements ListFieldCallback {

        private final FileBrowserListener listener;

        private Vector                   listItems = new Vector();

        public FileBrowserList() {

            // The number of rows is dynamic and it depends on the number of
            // files in the current directory, but the list is populated later
            super(0, Field.FIELD_TOP);

            changeDirectory(0);

            setEditable(false);
            setEmptyString("", DrawStyle.LEFT);

            setCallback(this);
            listener = new FileBrowserListener();
            addListeners();
        }

        public void changeDirectory(int index) {
            int count = 0;
            // Show directories first
            Enumeration dirs = getDirectoryEntries();
            while(dirs.hasMoreElements()) {
                String dir = (String)dirs.nextElement();
                if (count == listItems.size()) {
                    // We need to add a new entry in the list
                    insert(0);
                    FileBrowserListItem item = new FileBrowserListItem(dir, folderIcon);
                    listItems.addElement(item);
                } else {
                    // An item existed already, we just need to refresh it
                    FileBrowserListItem oldItem = (FileBrowserListItem)listItems.elementAt(count);
                    oldItem.setValue(dir);
                    oldItem.setIcon(folderIcon);
                    oldItem.setSeparator(false);
                    oldItem.setSelected(false);
                }
                count++;
            }
            // Now add the separator
            if (count == listItems.size()) {
                insert(0);
                FileBrowserListItem separator = new FileBrowserListItem(null, null);
                separator.setSeparator(true);
                listItems.addElement(separator);
            } else {
                FileBrowserListItem oldItem = (FileBrowserListItem)listItems.elementAt(count);
                oldItem.setValue(null);
                oldItem.setIcon(null);
                oldItem.setSeparator(true);
            }
            count++;
            // Add the files
            Enumeration files = getFileEntries();
            while(files.hasMoreElements()) {
                String file = (String)files.nextElement();
                if (count == listItems.size()) {
                    // We need to add a new entry in the list
                    insert(0);
                    FileBrowserListItem item = new FileBrowserListItem(file, null);
                    listItems.addElement(item);
                } else {
                    // An item existed already, we just need to refresh it
                    FileBrowserListItem oldItem = (FileBrowserListItem)listItems.elementAt(count);
                    oldItem.setValue(file);
                    oldItem.setIcon(null);
                    oldItem.setSeparator(false);
                    oldItem.setSelected(count == index);
                }
                count++;
            }
            // Now remove any left items
            int itemsToDelete = listItems.size() - count;
            for(int i=0;i<itemsToDelete;++i) {
                // We must always remove the last one
                int remIdx = listItems.size() - 1;
                delete(remIdx);
                listItems.removeElementAt(remIdx);
            }
            // We select the index if there is at least one (real) item
            if (count > 1) {
                setSelectedIndex(index);
            }
            // Set the title
            resetTitle();
            // Forces a repaint
            invalidate();
        }

        private int numRows() {
            return listItems.size();
        }

        // ListFieldCallback functions
        public void drawListRow(ListField listField, Graphics graphics, int index, int y, int width) {
            FileBrowserListItem item = ((FileBrowserListItem) listItems.elementAt(index));
            item.draw(graphics, 0, y, width, listField.getRowHeight());
        }

        public Object get(ListField listField, int index) {
            return null;
        }

        public String getSelectedEntry() {
            if (listItems.size() == 0) {
                return "";
            }

            int current = getSelectedIndex();
            if (current >= listItems.size()) {
                // In this case there is something wrong, but we better avoid
                // an ArrayIndexOutOfBoundException
                current = listItems.size() - 1;
            }
            FileBrowserListItem item = ((FileBrowserListItem) listItems.elementAt(current));
            return item.getValue(); 
        }

        public int indexOfList(ListField listField, String prefix, int start) {
            return listField.getSelectedIndex();
        }

        public int getPreferredWidth(ListField listField) {
            return getContentWidth();
        }

        protected int moveFocus(int amount, int status, int time) {
            FileBrowserListItem item;

            int oldSelection = getSelectedIndex();
            item = ((FileBrowserListItem) listItems.elementAt(oldSelection));
            item.setSelected(false);

            // Forward the call
            int ret = super.moveFocus(amount, status, time);

            int newSelection = getSelectedIndex();

            // Select the new item
            item = ((FileBrowserListItem) listItems.elementAt(newSelection));
            if (item.isSeparator()) {
                // We cannot move on a separator and shall skip it
                if (newSelection > oldSelection) {
                    // Moving downward
                    if (newSelection + 1 < listField.numRows()) {
                        newSelection++;
                    } else {
                        newSelection = oldSelection;
                    }
                } else {
                    // Moving upward
                    if (newSelection - 1 >= 0) {
                        newSelection--;
                    } else {
                        newSelection = oldSelection;
                    }
                }
            }
            item = ((FileBrowserListItem) listItems.elementAt(newSelection));
            item.setSelected(true);
            setSelectedIndex(newSelection);

            invalidate();

            return ret;
        }

        public void addListeners() {
            addKeyListener(listener);
            addTrackwheelListener(listener);
        }

        public void removeListeners() {
            removeKeyListener(listener);
            removeTrackwheelListener(listener);
        }

        public void setSelectedIndex(int index) {
            int current = getSelectedIndex();
            super.setSelectedIndex(index);
            ((FileBrowserListItem) listItems.elementAt(current)).setSelected(false);
            ((FileBrowserListItem) listItems.elementAt(index)).setSelected(true);
            invalidate();
        }
    }

    private class FileBrowserListItem {

        public static final int MARGIN_SIDE = 5;

        public static final int PADDING     = 2;

        private String          value;

        private Bitmap          icon;

        private boolean         selected    = false;

        private boolean         separator   = false;

        public FileBrowserListItem(String value, Bitmap icon) {
            this.value = value;
            this.icon = icon;
        }

        public void draw(Graphics graphics, int x, int y, int width, int height) {
            Log.trace("[FileBrowserListItem.draw]");
            if (separator) {
                drawSeparator(graphics, x, y, width, height);
            } else {
                drawBackground(graphics, x, y, width, height);
                int leftImageWidth = drawIcon(graphics, x, y, height);
                drawName(graphics, leftImageWidth, y, width, height);
            }
        }

        private void drawName(Graphics graphics, int x, int y, int width, int height) {

            if (selected) {
                graphics.setColor(Color.WHITE);
            } else {
                graphics.setColor(Color.BLACK);
            }

            graphics.drawText(value, x + PADDING + 3, y + PADDING + 2, DrawStyle.LEFT
                                    | DrawStyle.TOP, width - x - (PADDING * 2));

        }

        private void drawSeparator(Graphics graphics, int x, int y, int width, int height) {

            graphics.setColor(Color.WHITE);
            graphics.fillRect(x - 1, y - 1, width + 2, height + 1);
            graphics.setColor(Color.BLACK);
            graphics.drawLine(x, y + (height / 2), x + width, y + (height / 2));
        }

        public void setSeparator(boolean separator) {
            this.separator = separator;
        }

        public boolean isSeparator() {
            return separator;
        }

        private int drawIcon(Graphics graphics, int x, int y, int height) {

            if (icon == null) {
                return 0;
            }

            int imageWidth  = icon.getWidth();
            int imageHeight = icon.getHeight();
            int imageTop = y + ((height - imageHeight) / 2);
            int imageLeft = x + ((height - imageWidth) / 2);
            graphics.drawBitmap(imageLeft, imageTop, imageWidth, imageHeight, icon, 0, 0);

            return height;
        }

        private void drawBackground(Graphics graphics, int x, int y, int width, int height) {
            if (selected) {
                graphics.setColor(Color.BLUE);
                graphics.fillRect(x - 1, y - 1, width + 2, height + 1);
            } else {
                graphics.setColor(Color.WHITE);
                graphics.fillRect(x - 1, y - 1, width + 2, height + 1);
            }
        }

        public void setSelected(boolean selected) {
            this.selected = selected;
        }

        public String getValue() {
            return value;
        }

        public void setValue(String value) {
            this.value = value;
        }

        public void setIcon(Bitmap icon) {
            this.icon = icon;
        }
    }

    private class FileBrowserMenu {

        public FileBrowserMenu() {
        }

        /**
         * The menu
         */
        private final MenuItem selectMenu =
            new MenuItem(StaticDataHelper.getLanguage("menu_select"), 20000, 10) {
                public void run() {
                    choose();
                }
            };

        private final MenuItem   cancelMenu  =
            new MenuItem(StaticDataHelper.getLanguage("menu_cancel"), 20000, 10) {
                public void run() {
                    quit();
                }
            };

        private final MenuItem   upMenu  =
            new MenuItem(StaticDataHelper.getLanguage("menu_up"), 20000, 10) {
                public void run() {
                    up();
                }
            };

        public void makeMenu(Menu menu, String entry)
        {
            if (!ROOT.equals(currDirName)) {
                menu.add(upMenu);
                // In SELECT_FILE_MODE user can only select if the currectly selected item
                // is a file
                boolean add = true;
                if (selectMode == SELECT_FILE_MODE) {
                    if (entry != null && isDirectory(entry)) {
                        add = false;
                    }
                }
                if (add) {
                    menu.add(selectMenu);
                }
            }
            menu.add(cancelMenu);
        }
    }

    private class DirInfo {
        private String path     = null;
        private int    selected = -1;

        public DirInfo(String path) {
            this.path = path;
        }

        public void setSelected(int selected) {
            this.selected = selected;
        }

        public int getSelected() {
            return selected;
        }

        public String getPath() {
            return path;
        }
    }
}
